@@ -2,12 +2,14 @@ from functools import lru_cache
 from io import BytesIO
 from json import dumps as json_dumps
 from json import loads as json_loads
+from os import stat
 from os.path import abspath
 from os.path import dirname
 from os.path import isfile
 from os.path import join
 from os.path import split
 from re import sub as re_sub
+from typing import Callable
 from typing import Dict
 from typing import List
 from typing import Optional
@@ -37,6 +39,7 @@ app: Flask = Flask(
     static_folder=join(abspath(dirname(__file__)), "static")
 )
 db_path: str = "FA.db"
+m_time: Callable[[str], float] = lambda f: stat(f).st_mtime
 
 
 def clean_username(username: str, exclude: str = "") -> str:
@@ -96,6 +99,29 @@ def load_user(username: str) -> Optional[dict]:
         return db.users[username]
 
 
+@lru_cache
+def load_user_stats(username: str, _cache=None) -> Dict[str, int]:
+    username = clean_username(username)
+    stats: Dict[str, int] = {}
+    with FADatabase(db_path) as db:
+        stats["gallery"] = db.submissions.select(
+            {"replace(lower(author), '_', '')": username, "folder": "gallery"}, ["count(ID)"]
+        ).fetchone()[0]
+        stats["scraps"] = db.submissions.select(
+            {"replace(lower(author), '_', '')": username, "folder": "scraps"}, ["count(ID)"]
+        ).fetchone()[0]
+        stats["favorites"] = db.submissions.select(
+            {"favorite": f"%{username}%"}, ["count(ID)"], like=True
+        ).fetchone()[0]
+        stats["mentions"] = db.submissions.select(
+            {"mentions": f"%{username}%"}, ["count(ID)"], like=True
+        ).fetchone()[0]
+        stats["journals"] = db.journals.select(
+            {"replace(lower(author), '_', '')": username}, ["count(ID)"]
+        ).fetchone()[0]
+    return stats
+
+
 @lru_cache
 def load_item(table: str, id_: int) -> Tuple[Dict[str, Union[str, int]], int, int]:
     global db_path
@@ -129,7 +155,7 @@ def load_submission_file(id_: int) -> Tuple[Optional[str], str]:
 
 
 @lru_cache
-def search_table(table: str, sort: str, order: str, params_serialised: str = "{}", all_query: bool = False):
+def search_table(table: str, sort: str, order: str, params_serialised: str = "{}", force: bool = False, _cache=None):
     global db_path
 
     cols_results: List[str] = []
@@ -150,25 +176,41 @@ def search_table(table: str, sort: str, order: str, params_serialised: str = "{}
     sort = col_id if not sort else sort
 
     with FADatabase(db_path) as db:
-        cols_table: List[str] = db[table].columns
+        db_table: FADatabaseTable = db[table]
+        cols_table: List[str] = db_table.columns
 
-        if not params and not all_query:
+        if not params and not force:
             return [], cols_table, cols_results, cols_list, col_id, sort, order
 
-        params = {k: vs for k, vs in params.items() if k in map(str.lower, cols_table)}
+        params = {k: vs for k, vs in params.items() if k in map(str.lower, cols_table + ["any", "sql"])}
 
+        if "sql" in params:
+            query: str = " or ".join(map(lambda q: f"({q})", params["sql"]))
+            return (
+                list(db_table.cursor_to_dict(
+                    db_table.select_sql(query, columns=cols_results, order=[f"{sort} {order}"]),
+                    cols_results)),
+                cols_table + ["any"],
+                cols_results,
+                cols_list,
+                col_id,
+                sort,
+                order
+            )
         if "author" in params:
             params["replace(author, '_', '')"] = list(map(lambda u: clean_username(u, "%_"), params["author"]))
             del params["author"]
         if "username" in params:
             params["username"] = list(map(lambda u: clean_username(u, "%_"), params["username"]))
+        if "any" in params:
+            params[f"({'||'.join(cols_table)})"] = params["any"]
+            del params["any"]
 
-        db_table: FADatabaseTable = db[table]
         return (
             list(db_table.cursor_to_dict(
                 db_table.select(params, cols_results, like=True, order=[f"{sort} {order}"]),
                 cols_results)),
-            cols_table,
+            cols_table + ["any"],
             cols_results,
             cols_list,
             col_id,
@@ -188,6 +230,7 @@ def response_minify(response):
 @lru_cache
 @app.route("/user/<username>")
 def user(username: str):
+    global db_path
     if username != (username_clean := clean_username(username)):
         return redirect(f"/user/{username_clean}")
 
@@ -199,16 +242,18 @@ def user(username: str):
             404
         )
 
+    user_stats: Dict[str, int] = load_user_stats(username, _cache=m_time(db_path))
+
     return render_template(
         "user.html",
         title=f"{app.name} · {username}",
         user=username,
         folders=list(filter(bool, user_entry["FOLDERS"].split(","))),
-        gallery_length=f.count(",") + 1 if (f := user_entry["GALLERY"]) else 0,
-        scraps_length=f.count(",") + 1 if (f := user_entry["SCRAPS"]) else 0,
-        favorites_length=f.count(",") + 1 if (f := user_entry["FAVORITES"]) else 0,
-        mentions_length=f.count(",") + 1 if (f := user_entry["MENTIONS"]) else 0,
-        journals_length=f.count(",") + 1 if (f := user_entry["JOURNALS"]) else 0
+        gallery_length=user_stats["gallery"],
+        scraps_length=user_stats["scraps"],
+        favorites_length=user_stats["favorites"],
+        mentions_length=user_stats["mentions"],
+        journals_length=user_stats["journals"],
     )
 
 
@@ -227,6 +272,22 @@ def search_default():
     return redirect(url_for("search", table="submissions", **{k: request.args.getlist(k) for k in request.args}))
 
 
+@app.route("/gallery/<username>")
+@app.route("/search/gallery/<username>/")
+def search_user_gallery(username: str):
+    return redirect(url_for(
+        "search", table="submissions",
+        **{**{k: request.args.getlist(k) for k in request.args}, "author": username, "folder": "gallery"}))
+
+
+@app.route("/scraps/<username>")
+@app.route("/search/scraps/<username>/")
+def search_user_scraps(username: str):
+    return redirect(url_for(
+        "search", table="submissions",
+        **{**{k: request.args.getlist(k) for k in request.args}, "author": username, "folder": "scraps"}))
+
+
 @app.route("/submissions/<username>/")
 @app.route("/search/submissions/<username>/")
 def search_user_submissions(username: str):
@@ -250,8 +311,18 @@ def search_user_favorites(username: str):
             "favorite": f"%{username}%"}))
 
 
+@app.route("/mentions/<username>")
+@app.route("/search/mentions/<username>/")
+def search_user_mentions(username: str):
+    return redirect(url_for(
+        "search", table="submissions", **{
+            **{k: request.args.getlist(k) for k in request.args},
+            "mentions": f"%{username}%"}))
+
+
 @app.route("/search/<string:table>/")
 def search(table: str):
+    global db_path
     table = table.lower()
 
     if table not in ("submissions", "journals", "users"):
@@ -282,7 +353,8 @@ def search(table: str):
         sort,
         order,
         json_dumps(params),
-        all_query=request.path.startswith("/browse/")
+        force=request.path.startswith("/browse/"),
+        _cache=m_time(db_path)
     )
 
     return render_template(
@@ -330,7 +402,7 @@ def journal(id_: int):
 @lru_cache(maxsize=10)
 @app.route("/journal/<int:id_>/zip/")
 @app.route("/journal/<int:id_>/zip/<filename>")
-def journal_zip(id_: int, filename: str = ""):
+def journal_zip(id_: int, filename: str = None):
     jrn, _, _ = load_item(journals_table, id_)
 
     if jrn is None:
@@ -343,7 +415,7 @@ def journal_zip(id_: int, filename: str = ""):
         z.writestr("metadata.json", json_dumps({k: v for k, v in jrn.items() if k != "CONTENT"}).encode())
 
     f_obj.seek(0)
-    return send_file(f_obj, mimetype="application/zip", attachment_filename=filename if filename else None)
+    return send_file(f_obj, mimetype="application/zip", attachment_filename=filename)
 
 
 @app.route("/full/<int:id_>/")
@@ -375,13 +447,13 @@ def submission(id_: int):
 @lru_cache(maxsize=10)
 @app.route("/submission/<int:id_>/file/")
 @app.route("/submission/<int:id_>/file/<filename>")
-def submission_file(id_: int, filename: str = ""):
+def submission_file(id_: int, filename: str = None):
     sub_ext, sub_dir = load_submission_file(id_)
 
     if sub_ext is None:
         return abort(404)
     elif isfile(path := join(sub_dir, f"submission.{sub_ext}")):
-        return send_file(path, attachment_filename=filename if filename else None)
+        return send_file(path, attachment_filename=filename)
     else:
         return abort(404)
 
@@ -393,8 +465,9 @@ def submission_file(id_: int, filename: str = ""):
 @app.route("/submission/<int:id_>/thumbnail/<int:x>/<string:filename>")
 @app.route("/submission/<int:id_>/thumbnail/<int:x>x<int:y>/")
 @app.route("/submission/<int:id_>/thumbnail/<int:x>x<int:y>/<string:filename>")
-def submission_thumbnail(id_: int, x: int = 150, y: int = 150, filename: str = None):
+def submission_thumbnail(id_: int, x: int = 150, y: int = None, filename: str = None):
     sub_ext, sub_dir = load_submission_file(id_)
+    y = x if y is None else y
 
     if sub_ext is None:
         return abort(404)
@@ -416,7 +489,7 @@ def submission_thumbnail(id_: int, x: int = 150, y: int = 150, filename: str = N
 @lru_cache(maxsize=10)
 @app.route("/submission/<int:id_>/zip/")
 @app.route("/submission/<int:id_>/zip/<filename>")
-def submission_zip(id_: int, filename: str = ""):
+def submission_zip(id_: int, filename: str = None):
     sub, _, _ = load_item(submissions_table, id_)
 
     if sub is None:
@@ -432,7 +505,7 @@ def submission_zip(id_: int, filename: str = ""):
         z.writestr("metadata.json", json_dumps({k: v for k, v in sub.items() if k != "DESCRIPTION"}).encode())
 
     f_obj.seek(0)
-    return send_file(f_obj, mimetype="application/zip", attachment_filename=filename if filename else None)
+    return send_file(f_obj, mimetype="application/zip", attachment_filename=filename)
 
 
 @lru_cache(maxsize=10)
